header {package edu.vub.crime.parser;}

class ParserImpl extends Parser;

options {
  k = 5;
  buildAST = true;
  defaultErrorHandler = false;
}

// CRIME Code consists of a list of facts and rules
program : expression_list { /* todo tranform into a vector<expression> */ } ;

expression_list!	
			: exp:expression exps:more_expressions[#exp] { 
				#expression_list = #exps; 
			};

more_expressions![AST exp]	
			: EOF { #more_expressions = #exp; }
			| expl:expression_list {	#exp.setNextSibling(#expl);
									#more_expressions = #exp; }
			;

expression!	: actl:action_list con:condition[#actl] { #expression = #con; }
            | PACKAGE nam:symbol {
            	#nam.setText("package");
            	#expression = #([PACK, "package"],#nam);} 
			;
			
condition![AST actl]	
			  // Check for unbound variables in facts is performed by the tree walker.
			: DOT { #condition = #([AGASS, "assert"], #actl); } 
			  // Check for variables in the rule-head which are not bound by the patterns is performed by the tree walker.
			| INFERS patl:pattern_list { #condition = #([AGRULE, "rule"], #actl, #([AGCOND, "condition"], #patl)); }
			;

action_list!	: act:action acts:more_actions[#act] {
				#action_list = #acts; }
			;

more_actions![AST act]
			: COMMA actl:action_list	{	#act.setNextSibling(#actl); 
										#more_actions = #act; }
			| 					  	{	#more_actions = #act; }
			;

action!		: nam:action_name attrs:optional_attributes {
				#nam.addChild(#attrs);
				#action = #nam; }
			;

action_name! : COLON anam:symbol {	#anam.setText("name");
								#action_name = #([AGEVNT, "event"], #anam); }
								
			| fnam:pattern_name {	#action_name = #fnam; }
			;
			
pattern_name!: sym:symbol posq:possible_qualification[#sym] {
				#pattern_name = #posq; }
			;
						
possible_qualification![AST sym]
			: ARROW nam:symbol {
				#sym.setText("space");
				#nam.setText("name");
				#possible_qualification = #([AGQUA, "public fact"], #sym, #nam); }
			| { #sym.setText("name");
				#possible_qualification = #([AGUNQ, "private fact"], #sym); }
			;

optional_attributes!
			: 
			| LPR attr:attribute attrs:more_attributes[#attr] {
				#optional_attributes = #attr;
			};

more_attributes![AST attr]
			: RPR { #more_attributes = #attr; }
			| COMMA attrl:attribute_list {	#attr.setNextSibling(#attrl);
										#more_attributes = #attr; }
			;

attribute_list!
			: attr:attribute attrs:more_attributes[#attr] {
				#attribute_list = #attr;
			};


pattern_list!: pat:pattern pats:more_patterns[#pat] {
				#pattern_list = #pats; }
			;

more_patterns![AST pat]
			: COMMA patl:pattern_list	{	#pat.setNextSibling(#patl); 
										#more_patterns = #pat; }
			| DOT				  	{	#more_patterns = #pat; }
			;

pattern!	: NOT pos_pattern:ppattern
				{#pattern = #([NEGATION,"not"]);
 				#pattern.addChild(#pos_pattern);}		

			| pos_pat:ppattern 
				{#pattern = #pos_pat;} 				
			;			

ppattern!	: FINDALL LPR vars:attribute COMMA  LPR qwery:pattern_list RPR COMMA cumulative:attribute RPR {				
					#ppattern = #([PRIMFA,"findAll"]);
					#ppattern.addChild(#vars);
					#ppattern.addChild(#qwery);
					#ppattern.addChild(#cumulative);
				}

			| BAGOFF LPR varsb:attribute COMMA  LPR qweryb:pattern_list RPR COMMA cumulativeb:attribute RPR {				
				#ppattern = #([PRIMBO,"bagof"]);
				#ppattern.addChild(#varsb);
				#ppattern.addChild(#qweryb);
				#ppattern.addChild(#cumulativeb);
				}		
				
			| LNG LPR varlist:attribute COMMA varlength:attribute RPR {				
				#ppattern = #([LENGTH,"length"]);
				#ppattern.addChild(#varlist);
				#ppattern.addChild(#varlength);				
				}								
				
			| nam:pattern_name attrs:optional_attributes {
				#nam.addChild(#attrs);
				#ppattern = #nam; }
			;	

//attribute!	: LPR sym:symbol val:value RPR { #attribute = #([ AGATTR, "attribute"], #sym, #val); };

attribute!	: val:value { #attribute = #val; }
			| QUESTION var:symbol {
				#var.setType(AGVAR); 
				#var.setText("variable");
				#attribute = #var; }
			;

value		:! nbr:NBR { #value = #([AGNBR,"number"],nbr); }
			|! frc:FRC { #value = #([AGFRC,"fraction"],frc); }
			|! txt:TXT { #value = #([AGTXT,"text"],txt); }
			| symbol
			;
         
symbol!		: var:NAM { #symbol = #([AGSYM,"symbol"], var); };

//----------------------------------------------------------------------------------------------
//
//  LEXER
//
//----------------------------------------------------------------------------------------------
class LexerImpl extends Lexer;

options {
  k = 3;
}

// Protected Scanner Tokens

// OUTPUT TOKENS
// These tokens are never produced by the scanner itself as they are protected. 
// However they are used to annotate the resulting ANTLR tree, so that the walker
// can easily produce the correct Java ATAbstractGrammar elements.
// Each token definition aligns the token with its printed representation.
// Statements
protected PRIMFA 	: "findall";
protected PRIMBO	: "bagof";
protected PACK		: "%package";
protected NEGATION  : "not";
protected LENGTH : "length";
protected TYPE		: "type";
protected AGLST		: "list";			
protected AGFACT		: "fact";
protected AGRULE		: "rule";
protected AGASS		: "assert";
protected AGCOND		: "condition";
  			
protected AGACT		: "action";
protected AGPAT		: "pattern";  			
protected AGATTR		: "attribute";
protected AGSYM		: "symbol";

// Naming nodes for actions, facts and patterns
protected AGQUA		: "spaceName => factName";
protected AGUNQ		: "factName";
protected AGEVNT		: ":eventName";

// Literals
protected AGVAR		: "variable";
protected AGNBR		: "number";
protected AGFRC		: "fraction";
protected AGTXT		: "text";

// auxiliary tokens for operators
protected AGCMP     : "symbol";
protected AGADD     : "symbol";
protected AGMUL     : "symbol";
protected AGPOW     : "symbol";

protected DIGIT: '0'..'9'
    ;
    
protected LETTER: ('a'..'z'|'A'..'Z' | '_' | '.' )
    ;

protected EXPONENT: ('e' | 'E')
	;
	
protected CMPCHAR: ('!' | '<' | '=' | '>' | '~' )
	;

QUESTION: '?'
	;
		
protected SIGN: ('+' | '-' )
	;
	
protected SCALE: EXPONENT (SIGN)? NBR
	;
	
DOT: '.'
	;
	
COMMA: ','
	;	
	
ARROW: "->"
	;
	
FINDALL: "findall"
	;	
	
BAGOFF: "bagof"
	;	
	
NOT : "not"	
	;

LNG : "length"
	;	
	
PACKAGE : "%package"
		;	

protected COLON: ':'
	;
	
protected INFERS: ":-"
	;

COLON_OR_INFERS
	: (COLON '-') => INFERS 	{ $setType(INFERS); }
	| COLON					{ $setType(COLON); }
	;
	
protected NBR: (DIGIT)+
	;

protected FRC: NBR (SCALE | DOT NBR (SCALE)?)
	;
	
NBR_OR_FRC options { paraphrase = "a number or fraction"; }: ( NBR EXPONENT ) => FRC  { $setType(FRC); }
          | ( NBR DOT ) => FRC       { $setType(FRC); }
          |   NBR                    { $setType(NBR); }
    ;

protected CMP: CMPCHAR (CMPCHAR)?
	;

NAM options { paraphrase = "a symbol"; }: LETTER (DIGIT | LETTER)*
	;
		
WHITESPACE options { paraphrase = "whitespace"; }: ('\t' |  ' ')
           { $setType(Token.SKIP); }
    ;
    
NEWLINE options { paraphrase = "newline"; }:  ( "\r\n" | '\r' | '\n')
          { newline(); 
            $setType(Token.SKIP); }
    ;
    
LPR options { paraphrase = "a left parenthesis"; }: '(';
RPR options { paraphrase = "a right parenthesis"; }: ')';

TXT options { paraphrase = "a text string"; }: '"' (ESC|~('"'|'\\'|'\n'|'\r'))* '"'
	;

    // Single-line comments
SL_COMMENT options { paraphrase = "a single-line comment"; }
	:	"//" WHITESPACE
		(~('\n'|'\r'))* ('\n'|'\r'('\n')?)
		{$setType(Token.SKIP); newline();}
	;

// multiple-line comments
ML_COMMENT options { paraphrase = "a multi-line comment"; }
	:	"/*"
		(	/*	'\r' '\n' can be matched in one alternative or by matching
				'\r' in one iteration and '\n' in another. I am trying to
				handle any flavor of newline that comes in, but the language
				that allows both "\r\n" and "\r" and "\n" to all be valid
				newline is ambiguous. Consequently, the resulting grammar
				must be ambiguous. I'm shutting this warning off.
			 */
			options {
				generateAmbigWarnings=false;
			}
		:
			{ LA(2)!='/' }? '*'
		|	'\r''\n'		{newline();}
		|	'\r'			{newline();}
		|	'\n'			{newline();}
		|	~('*'|'\n'|'\r')
		)*
		"*/"
		{$setType(Token.SKIP);}
	;
	
protected ESC
	:	'\\'
		(	'n'
		|	'r'
		|	't'
		|	'b'
		|	'f'
		|	'"'
		|	'\''
		|	'\\'
		|	'0'..'3'
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:	'0'..'7'
				(
					options {
						warnWhenFollowAmbig = false;
					}
				:	'0'..'7'
				)?
			)?
		|	'4'..'7'
			(
				options {
					warnWhenFollowAmbig = false;
				}
			:	'0'..'7'
			)?
		)
	;

{	
	import edu.vub.crime.grammar.*;
	import java.util.LinkedList;
	import java.util.Set;
	import java.util.Iterator;
}
class TreeWalkerImpl extends TreeParser;

program returns [ ExpressionList result ] throws CrimeParseException
	{
		result = null;
		ExpressionList program = ExpressionList.whichAcceptsFactsAndRules();
	}
          : expression_list[program]
            { result = program; }
          ;

expression_list[ ExpressionList program ] throws CrimeParseException
	{
		Expression expr;
	}
			: ( expr=expression { program.add( expr ); } )+
			;

expression returns [ Expression expr ] throws CrimeParseException
	{	
		expr = null; 
	}
			: expr=rule
			| expr=pack
			| {	ExpressionList facts = ExpressionList.whichAcceptsFacts(); }
				
			  	#(AGASS pattern_list[facts])
				
			  {	if(facts.variables().isEmpty()) {
			  		expr = facts; 
			  	} else {
			  		throw new CrimeParseException("Asserted facts may never contain variables");
			  	}
			  }
			;

pack returns [Expression expr]
{
	expr = null;
	String pack=null;
}
	:  #(PACK pack=symbol)
		{ System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
			expr = ExpressionFactory.packageClass(pack); }
	;	

rule returns [ Expression expr ] throws CrimeParseException
	{
		expr = null;
		ExpressionList effect = ExpressionList.whichAcceptsFactsAndEvents();
		ExpressionList prerequisites = ExpressionList.whichAcceptsFactsAndPrimitives(); 
	}
			: #(AGRULE pattern_list[effect] #(AGCOND pattern_list[prerequisites]) )
			
			  {				  				  	
			  	if( prerequisites.variables().containsAll( effect.variables() )) {
			  		ExpressionList events = ExpressionList.whichAcceptsFactsAndEvents();
			  		for(Iterator itr = effect.iterator(); itr.hasNext();) {			  			
			  			Event event = ((Expression)itr.next()).makeEvent();
			  			events.add(event);
			  		}
			  		expr = ExpressionFactory.rule(events, prerequisites);
			  	} else {
			  		//Warning effect variables are mutated !!
			  		effect.variables().removeAll(prerequisites.variables());
			  		throw new RuntimeException("Effect of a rule contains unbound variables "+ effect.variables());
			  	}
			  }
			;

pattern_list[ ExpressionList patterns ] throws CrimeParseException
	{	
		Expression expr;
	}
			: ( expr=pattern { patterns.add( expr ); })*
			;
			

pattern returns [ Expression pattern ] throws CrimeParseException
	{pattern = null;}
	: {Expression negation,pat;}
		#(NEGATION pat=ppattern){
		negation  = ExpressionFactory.negation(pat);		
		pattern = negation;
	}
	| pattern = ppattern
	;
			

ppattern returns [ Expression pattern ] throws CrimeParseException
	{	
		pattern = null;
		String name;
	}
			: {	String space; }
				
				#(AGQUA 
					space=symbol 
					name=symbol 
					{	pattern = ExpressionFactory.qualifiedFact(space, name); }
					attributes[pattern])	
					
			|   #(PRIMFA 
					{Attribute vars, acum;			
				 	ExpressionList qwery = ExpressionList.whichAcceptsFacts();}
					vars=attribute
					pattern_list[qwery]
					acum=attribute
					{pattern = ExpressionFactory.findAll(vars,qwery,(Variable)acum);}
				)
				
			|   #(PRIMBO
					{Attribute vars, acum;			
				 	ExpressionList qwery = ExpressionList.whichAcceptsFacts();}
					vars=attribute
					pattern_list[qwery]
					acum=attribute
					{pattern = ExpressionFactory.bagOff(vars,qwery,(Variable)acum);}
				)
				
			|   #(LENGTH
					{Attribute varList, varLength;}
					varList=attribute					
					varLength=attribute
					{pattern = ExpressionFactory.length(varList,varLength);}
				)
												
			|	#(AGUNQ 
					name=symbol 
					{	pattern = ExpressionFactory.unqualifiedFact(name); }
						attributes[pattern])		
					
			|	#(AGEVNT
					name=symbol 
					{	pattern = ExpressionFactory.event(name); }
					attributes[pattern])
			;
			
attributes[ Expression fact_or_event ]
	{
		Attribute attr;
	}
			: ( attr=attribute { fact_or_event.addAttribute( attr ); })*
			;

attribute returns [ Attribute attr ]
	{	
		attr = null;
	}
			: attr=literal
			| attr=variable
			;

literal returns [ Attribute lit]
	{
		lit = null;
  	}
			: #(AGNBR nbr:NBR) { lit = ExpressionFactory.attribute(Integer.parseInt(nbr.getText())); }
			| #(AGFRC frc:FRC) { lit = ExpressionFactory.attribute(Double.parseDouble(frc.getText())); }
			| #(AGTXT txt:TXT) { String text = txt.getText(); lit = ExpressionFactory.attribute(text.substring(1, text.length() - 1)); }
			| {	String sym; }			
      			sym=symbol      			
      		  { lit = ExpressionFactory.symbol(sym); }
          ;
          
symbol returns [ String sym ]
	{	
		sym = null; 
	}
          : #(AGSYM txt:NAM) { sym = txt.getText(); }
          ;

variable returns [ Attribute var ]
	{	
		var = null; 
	}
          : #(AGVAR txt:NAM) { var = ExpressionFactory.variable(txt.getText()); }
          ;